// virtual disk integrity check utility
// Stewart Orchard Sept '98

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <dir.h>

typedef unsigned int uint;

enum buf_state_type {BUF_CLEAN, BUF_DIRTY};

// global variables

FILE *fp = NULL;

uint num_tracks, num_sides, dos_tracks, dos_sectors,
		bam_size, max_lsn, dir_lsn, bdir_lsn;

char *secbuf = NULL, *bam = NULL;
long current_sec = -1;
buf_state_type buf_state = BUF_CLEAN;

// structures representing directory slots

struct Clust_ptr_type   //storage pointer
{
	int start;
	char size;
};

struct Hdr_type         //directory header entry
{
	char flags;
	char fname[8];
	char ext[3];
	Clust_ptr_type ptr[4];
	char link;
};

struct Cont_type        //directory continuation entry
{
	char flags;
	Clust_ptr_type ptr[7];
	char pad[2];
	char link;
};

union Slot_type         //union holds either slot type
{
	Hdr_type hdr;
	Cont_type cont;
};

// errorlevel return codes

enum {ERR_OK, ERR_ALLOC, ERR_INFILE, ERR_OUTFILE, ERR_FORMAT,
		 ERR_NOTDRAGON};

// function prototypes

void load_vsector(uint trk, uint sec);
void quit(char *message, uint errlev);
void load_bam(void);
void load_dir(Slot_type *slot, uint slotnum);
void print_fname(Slot_type *slot);
void get_pointer(Slot_type &slot, uint &i, uint &lsn, uint &size);
void check_bam(void);
void check_cluster(uint lsn, uint size);
void find_lost_sectors(void);
void check_dir(void);
void check_file_sectors(void);
int check_cluster_sectors(uint lsn, uint size);
void check_xlinks(void);
int check_xlink_cluster(uint chk_lsn, uint chk_size, uint slotnum);


// repair function prototypes

void fix_file_past_end(uint slotnum, int &dir_end);
void write_dir(Slot_type &slot, uint slotnum);
void fix_nameless_file(uint slotnum);
void fix_bad_chars(uint slotnum);

//
//  M A I N   F U N C T I O N
//

int main(int argc, char *argv[])
{
	char fname[MAXPATH];
	char hdr[8];

	if((secbuf = new char[256]) == NULL)
	{
		quit("failed trying to allocate sector buffer", ERR_ALLOC);
	}

	if(argc > 1)
	{
		char drive[MAXDRIVE];
		char dir[MAXDIR];
		char file[MAXFILE];
		char ext[MAXEXT];
		strcpy(fname, argv[1]);
		if(!(fnsplit(fname, drive, dir, file, ext) & EXTENSION))
			strcpy(ext, "vdk");
		fnmerge(fname, drive, dir, file, ext);

		if((fp = fopen(fname, "r+b")) == NULL)
		{
			quit("Unable to open file", ERR_INFILE);
		}
	}
	else
	{
		quit("Virtual disk file must be specified", ERR_INFILE);
	}

	printf("Checking '%s'\n", fname);

	rewind(fp);
	fread(hdr, 8, 1, fp);

	if((hdr[0] != 'D') || (hdr[1] != 'K'))
	{
		quit("not a virtual disk file", ERR_FORMAT);
	}

	num_tracks = hdr[2];
	num_sides = hdr[3];

	if(((num_tracks != 80) && (num_tracks !=40))
	  || ((num_sides !=1) && (num_sides !=2)))
	{
		quit("not a virtual disk file or header corrupted", ERR_FORMAT);
	}

	long flen;
	fseek(fp, 0L, SEEK_END);
	flen = ftell(fp) - 8;

	if(flen != 256L*18*num_tracks*num_sides)
	{
		printf("Header contains incorrect information about virtual disk size\n");
	}

	printf("Virtual disk size: %d tracks, %d side(s)\n", num_tracks, num_sides);

	load_bam();
	printf("Dragon Dos formatted to: %d tracks, %d sectors per track\n", dos_tracks, dos_sectors);

	if(num_tracks * num_sides * 18 > max_lsn)
	{
		printf("\nThe virtual disk is larger than necessary\n");
	}

	printf("\nChecking directory...\n");
	check_dir();

	printf("Checking BAM (Block Availability Map)...\n");
	check_bam();
	find_lost_sectors();

	printf("Checking for cross-linked files...\n");
	check_xlinks();

	printf("\nChecking files for bad sectors...\n");
	check_file_sectors();

	quit("\nChecks complete", ERR_OK);

	return ERR_OK;	//suppresses return warning
}

//
// E N D  M A I N
//


// general exit function

void quit(char *message, uint errlev)
{
	if(buf_state == BUF_DIRTY)
	{
		fseek(fp, (current_sec<<8) + 8, SEEK_SET);
		fwrite(secbuf, 256, 1, fp);
	}

	printf("%s\n", message);

	if(fp != NULL) fclose(fp);
	if(secbuf != NULL) delete secbuf;
	if(bam != NULL) delete bam;

	exit(errlev);
}


// load virtual sector (if not already loaded)

void load_vsector(uint trk, uint sec)
{
	long secnum = trk*num_sides*18+sec-1;

	if(current_sec == secnum) return;

	if(buf_state == BUF_DIRTY)
	{
		fseek(fp, (current_sec<<8) + 8, SEEK_SET);
		fwrite(secbuf, 256, 1, fp);
	}

	fseek(fp, (secnum<<8) + 8, SEEK_SET);
	fread(secbuf, 256, 1, fp);
	current_sec = secnum;
	buf_state = BUF_CLEAN;
}

// load Dragon Dos BAM

void load_bam(void)
{
	load_vsector(20, 1);

	if( ((dos_tracks = secbuf[252]) != (~secbuf[254] & 0xff))
		|| ((dos_sectors = secbuf[253]) != (~secbuf[255] & 0xff)) )
	{
		dos_tracks = 80;
		dos_sectors = 36;
//		quit("not Dragon Dos format", ERR_NOTDRAGON);
	}

	dir_lsn = 20 * dos_sectors;
	bdir_lsn = 16 * dos_sectors;
	max_lsn = dos_tracks * dos_sectors;
	bam_size = max_lsn >> 3;

	if((bam = new char[bam_size]) == NULL)
	{
		quit("failed trying to allocate BAM buffer", ERR_ALLOC);
	}

	if(bam_size <= 180)
	{
		memcpy(bam, secbuf, bam_size);
	}
	else
	{
		memcpy(bam, secbuf, 180);
		load_vsector(20, 2);
		memcpy(&bam[180], secbuf, 180);
	}
}

// load a directory slot

void load_dir(Slot_type *slot, uint slotnum)
{
	uint secnum = slotnum / 10;
	uint offset = (slotnum - secnum*10)*25;
	uint i, t;

	load_vsector(20, secnum+3);
	memcpy(slot, &secbuf[offset], 25);

	//convert hi:lo pointers to lo:hi
	if(slot->hdr.flags & 1)				//continuation entry (7 pointers)
	{
		for(i=0; i<7; i++)
		{
			t = slot->cont.ptr[i].start;
			slot->cont.ptr[i].start = (t>>8) | (t<<8);
		}
	}
	else
	{
		for(i=0; i<4; i++)              //header entry (4 pointers)
		{
			t = slot->hdr.ptr[i].start;
			slot->hdr.ptr[i].start = (t>>8) | (t<<8);
		}
	}
}


// get storage pointer from directory slot

void get_pointer(Slot_type &slot, uint &i, uint &lsn, uint &size)
{
	if(slot.hdr.flags & 1)	//continuation entry (7 pointers)
	{
		if(i>=7)
		{
			i=0;
			if(slot.hdr.flags & 0x20)
			{
				load_dir(&slot, slot.hdr.link);
			}
			else
			{
				size = 0;
				return;
			}
		}

		size = slot.cont.ptr[i].size;
		lsn = slot.cont.ptr[i].start;
	}
	else  //header entry (4 pointers)
	{
		if(i>=4)
		{
			i=0;
			if(slot.hdr.flags & 0x20)
			{
				load_dir(&slot, slot.hdr.link);
				size = slot.cont.ptr[i].size;
				lsn = slot.cont.ptr[i].start;
			}
			else
			{
				size = 0;
			}
			return;
		}
		size = slot.hdr.ptr[i].size;
		lsn = slot.hdr.ptr[i].start;
	}
}


// print filename

void print_fname(Slot_type *slot)
{
	char ch;

	for(int i=0; i<11; i++)
	{
		ch = slot->hdr.fname[i];
		if(!ch) if(i<8) {i=7; continue;	} else break;
		if(i==8) putch('.');
		putch(ch);
	}
}

// check files against BAM

void check_bam(void)
{
	uint i, slotnum=0, lsn, size;

	Slot_type slot;

	check_cluster(20*dos_sectors, 18);	//main directory track
	check_cluster(16*dos_sectors, 18);  //backup directory track

	do
	{
		load_dir(&slot, slotnum++);

		if(slot.hdr.flags & 0x08) break;
		if(slot.hdr.flags & 0x81) continue;

		print_fname(&slot);

		for(i=0; ; i++)
		{
			get_pointer(slot, i, lsn, size);
			if(!size) break;
			printf(" %dx%d *", lsn, size);
			check_cluster(lsn, size);
		}

		printf("\t");

	} while(slotnum<160);
}


// check cluster against BAM

void check_cluster(uint lsn, uint size)
{
	uint offset;
	char bit;

	for( ; size; lsn++, size--)
	{
		offset = lsn>>3;
		bit = 1 << (lsn & 7);
		if(bam[offset] & bit)
		{
			printf("%d,", lsn);
		}

		bam[offset] |= bit; // free up sector for lost sector check
	}
}

// find lost sectors (after check_cluster has set BAM bits)

void find_lost_sectors(void)
{
	uint bit, offset, lsn = 0, lost_sectors = 0, consec = 0, chains = 0;

	for(offset=0; offset<bam_size; offset++)
		for(bit=1; bit<256; lsn++, bit*=2)
		{
			if(!(bam[offset] & bit))
			{
				lost_sectors++;
				if(!consec)
				{
					printf("lost chain starting at LSN %d ", lsn);
					chains++;
				}
				consec++;
			}
			else
			{
				if(consec) printf("& %d sectors long\n", consec);
				consec = 0;
			}
		}

	if(consec) printf("& %d sectors long\n", consec);

	if(lost_sectors)
		printf("%d lost sectors in %d chains\n", lost_sectors, chains);
}

// check directory entries

void check_dir(void)
{
	uint i, ch, lsn, size, slotnum = 0, errors, num_files=0;
	int dir_end =-1;
	Slot_type slot;

	do
	{
		load_dir(&slot, slotnum);

		if((slot.hdr.flags & 0x08) && dir_end == -1) dir_end = slotnum;
		if(slot.hdr.flags & 0x81) continue;

		num_files++;

		//check filename for illegal characters
		if(*slot.hdr.fname==0)
		{
			printf("File found with no filename. %d\n", slotnum);
			fix_nameless_file(slotnum);
		}

		errors = 0;

		for(i=0; i<11; i++)
		{
			ch = slot.hdr.fname[i];
			if(!ch)	if(i<8) {i=7; continue;	} else break;
			if(!isalpha(ch) && !isdigit(ch) && (ch != '-')) errors++;
		}
		if(errors)
		{
			print_fname(&slot);
			printf(" contains illegal characters.\n");
			fix_bad_chars(slotnum);
		}

		//check storage pointers
		errors = 0;

		for(i=0; ; i++)
		{
			get_pointer(slot, i, lsn, size);
			if(!size) break;
			if ((lsn+size) > max_lsn) errors++;
		}

		if(errors)
		{
			load_dir(&slot, slotnum);
			print_fname(&slot);
			printf(" has a problem with its storage pointers.\n");
		}

		//possibly valid file is beyond effective end of directory
		if(dir_end != -1)
		{
			printf("File '");
			print_fname(&slot);
			printf("' found beyond end of directory. %d\n", slotnum);
			fix_file_past_end(slotnum, dir_end);
		}

	} while(++slotnum<160);

	printf("\n%d files in directory\n", num_files);
}

// check files for bad sectors

void check_file_sectors(void)
{
	uint i, lsn, size, slotnum = 0, errors;
	Slot_type slot;

	do
	{
		load_dir(&slot, slotnum);

		if(slot.hdr.flags & 0x08) break;
		if(slot.hdr.flags & 0x81) continue;

		errors = 0;
		for(i=0; ; i++)
		{
			get_pointer(slot, i, lsn, size);
			if(!size) break;
			errors+= check_cluster_sectors(lsn, size);
		}

		if(errors)
		{
			load_dir(&slot, slotnum);
			print_fname(&slot);
			printf(" contains %d bad sectors.\n", errors);
		}

	} while(++slotnum<160);
}


// check a cluster for bad sectors

int check_cluster_sectors(uint lsn, uint size)
{
	uint i, track, sector, errcnt=0;

	for(i=0; i<size; lsn++, i++)
	{
		track = lsn / dos_sectors;
		sector = lsn - (track * dos_sectors) + 1;
		load_vsector(track, sector);
		if(!memcmp(secbuf, "?*!BaD S3CT0R?*!", 16)) errcnt++;
	}

	return errcnt;
}

// check for cross-linked files

void check_xlinks(void)
{
	uint i, lsn, size, slotnum=0;
	int errslot=-1;
	Slot_type slot;

	do
	{
		load_dir(&slot, slotnum);

		if(slot.hdr.flags & 0x08) break;
		if(slot.hdr.flags & 0x81) continue;

		for(i=0; ; i++)
		{
			get_pointer(slot, i, lsn, size);
			if(!size) break;

			if((errslot = check_xlink_cluster(lsn, size, slotnum+1)) != -1 )
				break;
		}

		if(errslot != -1)
		{
			load_dir(&slot, slotnum);
			print_fname(&slot);
			if(errslot == dir_lsn)
			{
				printf(" overlaps directory track");
			}
			else if(errslot == bdir_lsn)
			{
				printf(" overlaps backup directory track");
			}
			else
			{
				printf(" cross-links with ");
				load_dir(&slot, errslot);
				print_fname(&slot);
			}
			printf("\n");
		}

	} while(++slotnum<160);
}

// check if cluster cross-linked with any others

int check_xlink_cluster(uint chk_lsn, uint chk_size, uint slotnum)
{
	uint i, lsn, size;
	Slot_type slot;

	if( ((chk_lsn >= dir_lsn) && (chk_lsn < (dir_lsn+18)))
	 || ((dir_lsn >= chk_lsn) && (dir_lsn < (chk_lsn+chk_size))) )
		return dir_lsn;

	if( ((chk_lsn >= bdir_lsn) && (chk_lsn < (bdir_lsn+18)))
	 || ((bdir_lsn >= chk_lsn) && (bdir_lsn < (chk_lsn+chk_size))) )
		return bdir_lsn;

	do
	{
		load_dir(&slot, slotnum);

		if(slot.hdr.flags & 0x08) break;
		if(slot.hdr.flags & 0x81) continue;

		for(i=0; ; i++)
		{
			get_pointer(slot, i, lsn, size);
			if(!size) break;

			if( ((lsn >= chk_lsn) && (lsn < (chk_lsn+chk_size)))
			 ||((chk_lsn >= lsn) && (chk_lsn < (lsn+size))) )
				return slotnum;
		}
	} while(++slotnum < 160);

	return -1;
}

//
// R E P A I R   F U N C T I O N S
//

void fix_file_past_end(uint slotnum, int &dir_end)
{
	Slot_type slot;

	load_dir(&slot, slotnum);
	slot.hdr.flags &= 0xf7;
	write_dir(slot, (uint)dir_end);

	load_dir(&slot, slotnum);
	slot.hdr.flags = 0x89;
	write_dir(slot, slotnum);
	dir_end++;
}


void write_dir(Slot_type &slot, uint slotnum)
{
	uint secnum = slotnum / 10;
	uint offset = (slotnum - secnum*10)*25;
	uint i, t;

	load_vsector(20, secnum+3);

	//convert lo:hi pointers to hi:lo
	if(slot.hdr.flags & 1)				//continuation entry (7 pointers)
	{
		for(i=0; i<7; i++)
		{
			t = slot.cont.ptr[i].start;
			slot.cont.ptr[i].start = (t>>8) | (t<<8);
		}
	}
	else
	{
		for(i=0; i<4; i++)              //header entry (4 pointers)
		{
			t = slot.hdr.ptr[i].start;
			slot.hdr.ptr[i].start = (t>>8) | (t<<8);
		}
	}

	memcpy(&secbuf[offset], &slot, 25);
	buf_state = BUF_DIRTY;
}


void fix_nameless_file(uint slotnum)
{
	static char count = 'A';

	Slot_type slot;

	load_dir(&slot, slotnum);

	if(slot.hdr.ptr[0].size)
	{
		strcpy(slot.hdr.fname, "NONAME");
		*slot.hdr.ext = count;
		printf("...renamed to '");
		print_fname(&slot);
		printf("'\n");
	}
	else
	{
		slot.hdr.flags = 0x81;
		printf("... file KILLED as it was of zero length\n");
	}
	write_dir(slot, slotnum);
}


void fix_bad_chars(uint slotnum)
{
	Slot_type slot;
	char ch, i;

	load_dir(&slot, slotnum);
	for(i=0; i<11; i++)
	{
		ch = slot.hdr.fname[i];
		if(!ch)	if(i<8) {i=7; continue;	} else break;
		if(!isalpha(ch) && !isdigit(ch) && (ch != '-')) slot.hdr.fname[i] = '-';
	}
	printf("...renamed to '");
	print_fname(&slot);
	printf("'\n");
	write_dir(slot, slotnum);
}